from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.dialects.postgres import UUID
from .base import (db)
from .crud import (CRUDMixin)


class EntryMixin:

    @property
    def entry_type(self):
        return self.__class__.__name__.lower()

    @property
    def entry_id(self):
        return self.id


class EntryColumnMixin(CRUDMixin):
    field = None

    @declared_attr
    def _entry_type(self):
        return db.Column('entry_type', db.String(255), index=True)

    @declared_attr
    def _entry_id(self):
        return db.Column('entry_id', UUID(as_uuid=True), index=True)

    @hybrid_property
    def entry_type(self):
        if self.has_property('entry_type'):
            return self._entry_type
        return self.__class__.__name__.lower()

    @entry_type.setter
    def entry_type(self, value):
        self._entry_type = value

    @entry_type.expression
    def entry_type(cls):
        return cls._entry_type

    @hybrid_property
    def entry_id(self):
        if self.has_property('entry_id'):
            return self._entry_id
        return self.id

    @entry_id.setter
    def entry_id(self, value):
        self._entry_id = value

    @entry_id.expression
    def entry_id(cls):
        return cls._entry_id

    @classmethod
    def for_entries(cls, items, field=None, child=None):
        field = field or cls.field

        if not field:
            raise AttributeError('需要设置field值')

        data = items
        if child:
            data = [getattr(item, child) for item in items]

        data_id = set([item.id for item in data])
        results = dict((item_id, False) for item_id in data_id)

        if current_user.is_authenticated:
            res = cls.query \
                .filter(cls.entry_id.in_(data_id)) \
                .filter_by(user_id=current_user.id) \
                .all()
            for item in res:
                results[item.entry_id] = True

        if field:
            for item in data:
                setattr(item, field, results.get(item.id))

        return items





from sqlalchemy.ext.declarative import declared_attr
from .base import db

class DeletedMixin(object):

    @declared_attr
    def deleted(self):
        return db.Column('deleted', db.Boolean, default=False, index=True, doc='是否删除')


from flask import render_template
from flask.views import View


class BaseView(View):
    model = None
    template = None
    per_page = 20
    order_by = None

    def get_model(self):
        """ 获取模型 """
        if not self.model:
            raise AttributeError('需要设置model值')
        return self.model

    def get_template(self):
        """ 获取模版 """
        if not self.template:
            raise AttributeError('需要设置template值')
        return self.template

    def get_stmt(self):
        """ 自定义查询语句 """
        return None

    def get_query_filter(self):
        """ 自定义查询条件 """
        return {}

    def get_query(self, **kwargs):
        """ 获取查询语句 """
        if self.get_stmt():
            return self.get_stmt()

        stmt = self.model.query

        # 处理url主体,?之前
        filter_by_ = {}

        kwargs.update(self.get_query_filter())

        for k, v in kwargs.items():
            if self.model.has_property(k):
                filter_by_[k] = v

        if filter_by_:
            stmt = stmt.filter_by(**filter_by_)

        # 处理url参数?之后
        # 默认参数为多值,使用in操作符
        # 如果参数为单值,并且开头或结尾为%,使用like操作符
        filter_ = []

        for k, v in request.args.lists():
            if self.model.has_property(k):
                if len(v) == 1 and (v[0].startswith('%') or v[0].endswith('%')):
                    filter_.append(getattr(self.model, k).ilike(v[0]))
                else:
                    filter_.append(getattr(self.model, k).in_(v))

        if filter_:
            stmt = stmt.filter(*filter_)

        if self.model.has_property('deleted'):
            stmt = stmt.filter_by(deleted=False)

        # 处理自定义排序
        if self.order_by and isinstance(self.order_by, (list, tuple)):
            stmt = stmt.order_by(*self.order_by)

        return stmt

    def get_context(self):
        """ 获取自定义上下文 """
        return {}

    def merge_data(self, items):
        """ 数据合并 """
        return items

    def render(self, **context):
        """ 渲染模版 """
        return render_template(self.get_template(), **context)


from flask import Blueprint as BaseBlueprint
import types


class Blueprint(BaseBlueprint):

    def route(self, rule, **options):

        def decorator(f):
            endpoint = options.pop('endpoint', f.__name__)

            view_func = f

            if not isinstance(f, types.FunctionType):
                class_args = options.pop('class_args', [])
                class_kwargs = options.pop('class_kwargs', {})
                view_func = f.as_view(endpoint, *class_args, **class_kwargs)

            self.add_url_rule(rule, endpoint, view_func, **options)
            return f

        return decorator


from flask import session
from flask_oauthlib.client import OAuth
from werkzeug.urls import url_parse, url_encode

oauth = OAuth()


weibo = oauth.remote_app(
    'weibo',
    request_token_params={'scope': 'email,statuses_to_me_read'},
    base_url='https://api.weibo.com/2/',
    authorize_url='https://api.weibo.com/oauth2/authorize',
    request_token_url=None,
    access_token_method='POST',
    access_token_url='https://api.weibo.com/oauth2/access_token',
    content_type='application/json',
    app_key='WEIBO',
)


@weibo.tokengetter
def get_weibo_oauth_token():
    return session.get('weibo_token')


def change_weibo_header(uri, headers, body):
    auth = headers.get('Authorization')
    if auth:
        auth = auth.replace('Bearer', 'OAuth2')
        headers['Authorization'] = auth
    return uri, headers, body

weibo.pre_request = change_weibo_header


qq = oauth.remote_app(
    'qq',
    base_url='https://graph.qq.com',
    request_token_url=None,
    request_token_params={'scope': 'get_user_info'},
    access_token_url='/oauth2.0/token',
    authorize_url='/oauth2.0/authorize',
    app_key='QQ',
)


@qq.tokengetter
def get_qq_oauth_token():
    return session.get('qq_token')


def change_qq_header(uri, headers, body):
    return uri, headers, body

qq.pre_request = change_qq_header


douban = oauth.remote_app(
    'douban',
    base_url='https://api.douban.com/',
    request_token_url=None,
    request_token_params={'scope': 'douban_basic_common,shuo_basic_r'},
    access_token_url='https://www.douban.com/service/auth2/token',
    authorize_url='https://www.douban.com/service/auth2/auth',
    access_token_method='POST',
    app_key='DOUBAN',
)


@douban.tokengetter
def get_douban_oauth_token():
    return session.get('douban_token')


weixin = oauth.remote_app(
    'weixin',
    app_key='WEIXIN',
    request_token_params={'scope': 'snsapi_base'},
    base_url='https://api.weixin.qq.com',
    authorize_url='https://open.weixin.qq.com/connect/qrconnect',
    access_token_params={'grant_type': 'authorization_code'},
    access_token_url='/sns/oauth2/access_token',
    content_type='application/json',
)


@weixin.tokengetter
def get_weixin_oauth_token():
    return session.get('weixin_token')


def weixin_oauth(weixin):
    """Fixes the nonstandard OAuth interface of Tencent WeChat."""

    original_methods = {
        'authorize': weixin.authorize,
        'authorized_response': weixin.authorized_response,
    }

    def authorize(*args, **kwargs):
        response = original_methods['authorize'](*args, **kwargs)
        url = url_parse(response.headers['Location'])
        args = url.decode_query()

        args['appid'] = args.pop('client_id')
        url = url.replace(query=url_encode(args), fragment='wechat_redirect')

        response.headers['Location'] = url.to_url()
        return response

    def authorized_response(*args, **kwargs):
        original_access_token_params = weixin.access_token_params
        weixin.access_token_params = {
            'appid': weixin.consumer_key,
            'secret': weixin.consumer_secret,
        }
        response = original_methods['authorized_response'](*args, **kwargs)
        weixin.access_token_params = original_access_token_params
        return response

    weixin.authorize = authorize
    weixin.authorized_response = authorized_response

    return weixin

weixin = weixin_oauth(weixin)



from flask_restful import fields


class FieldMixin(fields.Raw):

    def __init__(self, prompt='', *args, **kwargs):
        self.prompt = prompt
        super().__init__(*args, **kwargs)


class Raw(FieldMixin, fields.Raw):
    pass


class Integer(FieldMixin, fields.Integer):
    pass


class String(FieldMixin, fields.String):
    pass


class Url(FieldMixin, fields.Url):

    def __init__(self, prompt='', *args, **kwargs):
        kwargs['absolute'] = True
        super().__init__(prompt, *args, **kwargs)


class Boolean(FieldMixin, fields.Boolean):
    pass


class FormattedString(FieldMixin, fields.FormattedString):
    pass


class Float(FieldMixin, fields.Float):
    pass


class Arbitrary(FieldMixin, fields.Arbitrary):
    pass


class DateTime(FieldMixin, fields.DateTime):

    def __init__(self, prompt='', *args, **kwargs):
        kwargs['dt_format'] = 'iso8601'
        super().__init__(prompt, *args, **kwargs)


class Fixed(FieldMixin, fields.Fixed):
    pass


class List(fields.List):
    pass


class Nested(fields.Nested):
    pass




from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import orm


db = SQLAlchemy()


class BaseMixin(object):
    """ Base mixin
    """
    __table_args__ = {'extend_existing': True}

    @classmethod
    def column_properties(cls):
        properties = []
        mapper = getattr(cls, '__mapper__')

        if mapper and hasattr(mapper, 'iterate_properties'):
            properties = [p.key.lstrip('_') for p in mapper.iterate_properties if isinstance(p, (orm.ColumnProperty,))]

        return properties

    def as_dict(self, include=None, exclude=None):
        """
        :param include: 需要显示的属性列表
        :param exclude: 需要排除的属性列表
        :return:
        """
        fields = [field.strip('_') for field in self.column_properties()]

        exportable_fields = (include or []) + fields
        exportable_fields = set(exportable_fields) - set(exclude or [])

        result = dict()
        for field in exportable_fields:
            value = getattr(self, field)
            if hasattr(value, '__call__'):
                value = value()
            result[field] = value

        return result

    @classmethod
    def has_property(cls, name):
        return name in cls.column_properties()


from sqlalchemy.ext.declarative import declared_attr
from itsdangerous import TimedJSONWebSignatureSerializer as Serializer
from werkzeug.security import generate_password_hash, check_password_hash
from flask import current_app
from .base import db
from .crud import CRUDMixin
from uuid import UUID


class AccountMixin(CRUDMixin):

    @declared_attr
    def email(self):
        return db.Column('email', db.VARCHAR(255), index=True, unique=True, doc='用户邮箱,用来做登陆帐号')

    @declared_attr
    def password_hash(self):
        return db.Column('password_hash', db.VARCHAR(512), doc='用户密码的hash值')

    @declared_attr
    def email_confirmed(self):
        return db.Column('email_confirmed', db.Boolean, default=False, doc='邮箱是否经过确认')

    @property
    def password(self):
        raise ValueError('no password')

    @password.setter
    def password(self, password):
        setattr(self, 'password_hash', generate_password_hash(password))

    def verify_password(self, password):
        return check_password_hash(self.password_hash, password)

    @classmethod
    def get_by_token(cls, token, _=None):
        query = getattr(cls, 'query')
        data = cls.load_token(token)
        if data:
            return query.filter_by(id=UUID(data.get('user_id'))).first()

    @classmethod
    def get_by_account(cls, username, password):
        query = getattr(cls, 'query')
        user = query.filter_by(email=username).first()
        if user and user.verify_password(password):
            return user

    def generate_token(self, expiration=3600):
        """ 生成token"""
        s = Serializer(current_app.config['SECRET_KEY'], expiration)
        return s.dumps({'user_id': self.id.hex})

    @classmethod
    def verify_token(cls, token):
        data = cls.load_token(token)
        return cls.query.filter_by(id=data.get('user_id'))

    @property
    def token(self):
        return self.generate_token(expiration=current_app.config.get('', 60 * 60 * 24 * 30))

    @staticmethod
    def load_token(token):
        s = Serializer(current_app.config['SECRET_KEY'])
        try:
            return s.loads(token)
        except:
            return {}

    def confirm(self, token):
        data = self.load_token(token)

        if data and data.get('user_id') != self.id.hex:
            return False

        self.update(email_confirmed=True)
        return True



try:
    from flask import has_app_context
    from celery import Celery as BaseCelery

    class Celery(BaseCelery):
        app = None

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.patch_task()

            if 'application' in kwargs:
                self.init_app(kwargs['application'])

        def patch_task(self):
            base_task = self.Task
            _celery = self

            class ContextTask(base_task):
                abstract = True

                def __call__(self, *args, **kwargs):
                    if has_app_context():
                        return base_task.__call__(self, *args, **kwargs)

                    with _celery.app.app_context():
                        return base_task.__call__(self, *args, **kwargs)

            setattr(self, 'Task', ContextTask)

        def init_app(self, app):
            self.app = app
            self.config_from_object(app.config)
except:
    class Celery:
        pass



from functools import wraps
from flask import request, current_app


def jsonp(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        callback = request.args.get('callback', False)
        if callback:
            data = str(f(*args, **kwargs))
            content = str(callback) + '(' + data + ')'
            mime_type = 'application/javascript'
            return current_app.response_class(content, mimetype=mime_type)
        else:
            return f(*args, **kwargs)
    return decorated_function



from flask import (request, url_for)

from flask.ext.zeus.api.fields import (List, Nested)


class BaseResource(object):
    # 模型
    # type: sqlalchemy obj 需要继承 CRUDMixin
    model = None
    model_fields = None
    url_parse = None

    # 创建数据使用的表单
    # type: wtforms obj
    create_form = None

    # 更新数据使用的表单
    # type: wtforms obj
    update_form = None

    # 删除数据使用的表单
    # type: wtforms obj
    delete_form = None

    can_read = True
    # 是否允许创建
    # type: bool
    can_create = False

    # 是否允许更新
    # type: bool
    can_update = False

    # 是否允许删除
    # type: bool
    can_delete = False

    # 是否允许返回空数据
    # type: bool
    can_empty = True

    # 是否显示分页
    # type: bool
    can_paginate = True

    # 页码
    # type: int
    default_page = 1

    # 每页数据个数
    # type: int
    default_per_page = 20

    # 是否生成包含域名的完整url
    # type: bool
    is_full_url = True

    # 自定义排序
    # type: list or tuple or set
    order_by = None

    # 是否开始csrf验证
    # type: bool
    csrf_enabled = True

    @property
    def cls_name(self):
        """ 视图类名称
        :return: str
        """
        return self.__class__.__name__

    def check_model(self):
        """ 检查模型是否设置
        :return:
        """
        assert self.model, 'API Class: {} model not set'.format(self.cls_name)

    def check_create_form(self):
        """ 检查创建表单是否设置
        :return:
        """
        assert self.create_form, 'API Class: {} create_form not set'.format(self.cls_name)

    def check_update_form(self):
        """ 检查更新表单是否设置
        :return:
        """
        assert self.update_form, 'API Class: {} update_form not set'.format(self.cls_name)

    def check_model_fields(self):
        """ 检查输出格式是否设置
        :return:
        """
        assert self.model_fields, 'API Class: {} model_fields not set'.format(self.cls_name)

    def get_stmt(self, **kwargs):
        """ 自定义stmt语句, 需重写, 提供给get方法使用
        :return: sqlalchemy query
        """
        return None

    def generate_stmt(self, **kwargs):
        """ 生成查询语句
        :param kwargs:
        :return: sqlalchemy query
        """
        # 如有自定义语句,直接返回
        if self.get_stmt(**kwargs):
            return self.get_stmt(**kwargs)

        stmt = self.model.query

        # 处理url主体,?之前
        filter_by_ = {}

        for k, v in kwargs.items():
            filter_by_[k] = v

        if filter_by_:
            stmt = stmt.filter_by(**filter_by_)

        # 处理url参数?之后
        # 默认参数为多值,使用in操作符
        # 如果参数为单值,并且开头或结尾为%,使用like操作符
        filter_ = []

        for k, v in request.args.lists():
            if self.model.has_property(k):
                if len(v) == 1 and (v[0].startswith('%') or v[0].endswith('%')):
                    filter_.append(getattr(self.model, k).ilike(v[0]))
                else:
                    filter_.append(getattr(self.model, k).in_(v))

        if filter_:
            stmt = stmt.filter(*filter_)

        # 处理自定义排序
        if self.order_by and isinstance(self.order_by, (list, tuple)):
            stmt = stmt.order_by(*self.order_by)

        if self.model.has_property('deleted'):
            stmt = stmt.filter_by(deleted=False)

        return stmt

    def generate_iter_pages(self, pages, per_page, **kwargs):
        """ 生成分页
        :param pages:
        :param per_page:
        :return: list
        """
        pages = list(pages)
        iter_pages = []

        for page in pages:
            if isinstance(page, int):
                iter_pages.append({
                    'page': page,
                    'url': self.generate_url(page, per_page, **kwargs)
                })
            else:
                iter_pages.append({
                    'page': '...',
                    'url': ''
                })

        return iter_pages

    def generate_url(self, page, per_page, **kwargs):
        """ 生成链接
        :param page:
        :param per_page:
        :return: str
        """
        return url_for(request.endpoint, page=page, per_page=per_page, _external=self.is_full_url, **kwargs)

    def merge_data(self, data):
        return data

    def format_return(self, fields):

        data = {}
        for name, field in fields.items():
            if isinstance(field, List):
                try:
                    data[name] = [self.format_return(field.container.nested)]
                except:
                    data[name] = [getattr(field.container, 'prompt', '')]
            elif isinstance(field, Nested):
                data[name] = self.format_return(field.nested)
            else:
                data[name] = getattr(field, 'prompt', '')

        return data

    def output_args(self, url_parse):
        args = []
        for item in url_parse.args:
            item_data = dict()
            item_data['name'] = item.name
            item_data['help'] = item.help
            item_data['type'] = item.type.__name__
            item_data['required'] = item.required

            if item.choices:
                item_data['choices'] = item.choices

            if item.default:
                item_data['default'] = item.default

            args.append(item_data)

        return args


from flask import request
from flask_restful import (Resource, marshal)
from flask_login import (login_required, current_user)
from collections import ChainMap
from .base import BaseResource
from .errors import *


class RestfulResource(BaseResource, Resource):
    """
    example:
        from app.models import Post
        from app.forms import PostCreateForm, PostUpdateForm
        from app.model_fields import post_fields

        @api.resource('/posts/', '/posts/<int:id>/')
        class PostAPI(ModelResource):
            model = Post
            create_form = PostCreateForm
            update_form = PostUpdateForm
            model_fields = post_fields
            can_create = True
            can_update = True
            can_delete = True
    """

    def get(self, **kwargs):
        """ 资源获取
        :param kwargs:
        :return: dict
        """
        if not self.can_read:
            raise ZeusMethodNotAllowed

        self.check_model()
        self.check_model_fields()

        stmt = self.generate_stmt(**kwargs)

        if kwargs.get('id'):
            item = stmt.first()
            if not item:
                raise ZeusNotFound
            return marshal(item, self.model_fields)

        if self.can_paginate:
            page = request.args.get('page', self.default_page, int) or self.default_page
            per_page = request.args.get('per_page', self.default_per_page, int) or self.default_per_page
            pagination = stmt.paginate(page, per_page, error_out=not self.can_empty)
            items = self.merge_data(pagination.items)
            return {
                'items': marshal(items, self.model_fields),
                'pagination': OrderedDict([
                    ('has_prev', pagination.has_next),
                    ('has_next', pagination.has_next),
                    ('prev_num', pagination.prev_num),
                    ('next_num', pagination.next_num),
                    ('prev_url', self.generate_url(pagination.prev_num, per_page, **kwargs) if pagination.has_prev else ''),
                    ('next_url', self.generate_url(pagination.next_num, per_page, **kwargs) if pagination.has_next else ''),
                    ('page', pagination.page),
                    ('per_page', per_page),
                    ('pages', pagination.pages),
                    ('total', pagination.total),
                    ('iter_pages', self.generate_iter_pages(pagination.iter_pages(), per_page, **kwargs)),
                ])
            }
        else:
            items = stmt.all()

            if not self.can_empty and not items:
                raise ZeusNotFound

            return marshal(items, self.model_fields)

    @login_required
    def post(self, **kwargs):
        """ 资源创建
        :param kwargs:
        :return: dict
        """
        if not self.can_create:
            raise ZeusMethodNotAllowed

        self.check_model()
        self.check_create_form()
        self.check_model_fields()

        if not self.model.has_property('user_id'):
            raise ZeusMethodNotAllowed

        form = self.create_form(csrf_enabled=self.csrf_enabled)

        for k, v in kwargs.items():
            field = getattr(form, k, None)
            if field:
                field.data = v

        if form.validate_on_submit():
            data = ChainMap(form.data, kwargs)
            item = self.model()

            for k, v in data.items():
                setattr(item, k, v)

            if self.model.has_property('user_id'):
                item.user_id = current_user.id

            item.save()
            return marshal(item, self.model_fields), 201

        raise ZeusBadRequest(details=form.errors)

    @login_required
    def put(self, **kwargs):
        """ 资源更新
        :param kwargs:
        :return: dict
        """
        if not self.can_update:
            raise ZeusMethodNotAllowed

        self.check_model()
        self.check_update_form()
        self.check_model_fields()

        if not kwargs or not self.model.has_property('user_id'):
            raise ZeusMethodNotAllowed

        stmt = self.generate_stmt(**kwargs)
        item = stmt.first()

        if not item:
            raise ZeusNotFound

        if item.user_id != current_user.id:
            raise ZeusUnauthorized

        form = self.update_form(csrf_enabled=self.csrf_enabled)

        for k, v in kwargs.items():
            field = getattr(form, k)
            if field:
                field.data = v

        if form.validate_on_submit():
            for k, v in form.data.items():
                setattr(item, k, v)
            item.save()
            return marshal(item, self.model_fields), 200

        raise ZeusBadRequest(details=form.errors)

    @login_required
    def delete(self, **kwargs):
        """ 资源删除
        :param kwargs:
        :return: None
        """
        if not self.can_delete:
            raise ZeusMethodNotAllowed

        self.check_model()

        if not kwargs or not self.model.has_property('user_id'):
            raise ZeusMethodNotAllowed

        if self.delete_form:
            form = self.delete_form()
            if not form.validate_on_submit():
                raise ZeusBadRequest(details=form.errors)

        stmt = self.generate_stmt(**kwargs)
        item = stmt.first()

        if not item:
            raise ZeusNotFound

        if item.user_id != current_user.id:
            raise ZeusUnauthorized

        item.delete()

        return {}, 204

    def options(self, **kwargs):
        allow = []
        data = {}

        if self.can_read:
            allow.append('GET')

            if self.url_parse:
                data['GET'] = {'args': self.output_args(self.url_parse)}

        if self.can_create:
            allow.append('POST')
            data['POST'] = self.create_form.fields(**kwargs)

        if self.can_update:
            allow.append('PUT')
            data['PUT'] = self.update_form.fields(**kwargs)

        if self.can_delete:
            allow.append('DELETE')

        if self.model_fields:
            data['return'] = self.format_return(self.model_fields)

        headers = {
            'Allow': ', '.join(allow),
        }
        return data, headers


from sqlalchemy import text
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.dialects.postgres import UUID
from datetime import datetime
from .base import (db, BaseMixin)


class CRUDMixin(BaseMixin):
    """ Basic CRUD mixin
    """

    @declared_attr
    def id(self):
        return db.Column('id', UUID(as_uuid=True), server_default=text('uuid_generate_v4()'), primary_key=True, doc='主键')

    @declared_attr
    def created_at(self):
        return db.Column('created_at', db.DateTime, default=datetime.now, index=True, nullable=False, doc='创建时间')

    @declared_attr
    def updated_at(self):
        return db.Column('updated_at', db.DateTime, default=datetime.now, index=True, nullable=False, doc='更新时间')

    @declared_attr
    def deleted(self):
        return db.Column('deleted', db.Boolean, default=False, index=True, doc='是否删除')

    @classmethod
    def get(cls, row_id):
        query = getattr(cls, 'query')
        return query.get(row_id)

    @classmethod
    def create(cls, commit=True, **kwargs):
        return cls(**kwargs).save(commit)

    def save(self, commit=True):
        db.session.add(self)

        if commit:
            db.session.commit()

        return self

    def delete(self, commit=True):
        if self.has_property('deleted'):
            setattr(self, 'deleted', True)
            db.session.add(self)
        else:
            db.session.delete(self)

        if commit:
            db.session.commit()

    def update(self, commit=True, **kwargs):
        return self._set_attributes(**kwargs).save(commit)

    def _set_attributes(self, **kwargs):
        for k, v in kwargs.items():

            if k.startswith('_'):
                raise ValueError('私有属性不允许被设置')

            if self.has_property(k):
                setattr(self, k, v)

        return self



from flask import Flask


def create_app(import_name, config, config_name='default', ext_list=None, bp_list=None, **kwargs):
    app = Flask(import_name, **kwargs)
    app_config = config.get(config_name)
    app.config.from_object(app_config)

    if hasattr(app_config, 'init_app'):
        app_config.init_app(app)

    if ext_list and isinstance(ext_list, (list, tuple)):
        for ext in ext_list:
            ext.init_app(app)

    if bp_list and isinstance(bp_list, (list, tuple)):
        for bp in bp_list:
            app.register_blueprint(bp)

    return app


from werkzeug.exceptions import *
from flask_restful import HTTPException
from collections import OrderedDict


class ZeusHTTPException(HTTPException):
    """
    code http状态码
    description 错误描述
    """

    def __init__(self, description=None, response=None, error_code=0, details=None):
        """
        :param description: 自定义错误描述
        :param response: 自定义响应数据
        :param error_code: 自定义错误代码
        :param details: 自定义错误详情
        :return:
        """
        self.status_code = self.code
        self.message = description or self.description
        self.error_code = error_code
        self.details = details or {}
        super(ZeusHTTPException, self).__init__(description, response)

    def as_dict(self):
        """
        :return: {
            'status_code': ...,
            'error_code': ...,
            'message': ...,
            'details': ...
        }
        """
        data = OrderedDict()
        for k in ['status_code', 'error_code', 'message', 'details']:
            data[k] = getattr(self, k)
        return data

    @property
    def data(self):
        return self.as_dict()


class ZeusBadRequest(BadRequest, ZeusHTTPException):
    """ 400
    """
    pass


class ZeusUnauthorized(Unauthorized, ZeusHTTPException):
    """ 401
    """
    pass


class ZeusNotFound(NotFound, ZeusHTTPException):
    """ 404
    """
    pass


class ZeusMethodNotAllowed(MethodNotAllowed, ZeusHTTPException):
    """ 405
    """
    pass


from flask import request
from .base import BaseView


class BaseListView(BaseView):
    per_page = 20
    error_out = False

    def get_paginate_args(self):
        """ 获取分页参数 """
        page = request.args.get('page', 1, type=int)
        per_page = request.args.get('per_page', self.per_page, type=int)
        return page, per_page


class ListView(BaseListView):

    def dispatch_request(self, **kwargs):
        stmt = self.get_query(**kwargs)
        pagination = stmt.paginate(*self.get_paginate_args(), error_out=self.error_out)
        context = self.get_context()
        items = self.merge_data(pagination.items)
        context.update({
            'items': items,
            'pagination': pagination
        })
        return self.render(**context)


from .base import BaseView


class DetailView(BaseView):

    def dispatch_request(self, **kwargs):
        stmt = self.get_query(**kwargs)
        item = stmt.first()
        context = self.get_context()
        context.update({
            'item': item
        })
        return self.render(**context)


from flask_restful import (Resource, marshal)
from flask_login import (login_required, current_user)
from .base import BaseResource
from .errors import *


class ToggleResource(BaseResource, Resource):

    @login_required
    def post(self, **kwargs):
        self.check_model()
        self.check_model_fields()

        if not self.model.has_property('user_id'):
            raise ZeusMethodNotAllowed

        stmt = self.generate_stmt(**kwargs)
        item = stmt.first()

        if not item:
            item = self.model()
            for k, v in kwargs.items():
                if self.model.has_property(k):
                    setattr(item, k, v)
            item.user_id = current_user.id
            item.save()
            return marshal(item, self.model_fields), 201

        item.delete()
        return {}, 204


from flask import (request, url_for, flash, redirect)
from flask_login import (login_required, current_user)
from .base import BaseView


class BaseFormView(BaseView):
    form = None
    csrf_enabled = False
    success_message = None
    methods = ['GET', 'POST']
    decorators = [login_required]

    def get_form(self):
        """ 获取 wtforms 表单"""
        if not self.form:
            raise AttributeError('需要设置form值')

        return self.form

    def get_next_url(self, **kwargs):
        """ 获取跳转链接 """
        return request.args.get('next') or url_for(request.endpoint, **kwargs)


class CreateView(BaseFormView):

    def dispatch_request(self, **kwargs):
        form = self.get_form()(csrf_enabled=self.csrf_enabled)

        if form.validate_on_submit():

            item = self.model()

            for k, v in form.data.iteritems():
                setattr(item, k, v)

            item = item.update(commit=False, **kwargs)

            if self.model.has_property('user_id'):
                item.user_id = current_user.id

            item.save()

            if self.success_message:
                flash(self.success_message, category='success')

            return redirect(self.get_next_url(**kwargs))

        context = self.get_context()
        context.update({
            'form': form
        })
        return self.render(**context)


class UpdateView(BaseFormView):

    def dispatch_request(self, **kwargs):

        stmt = self.get_query(**kwargs)
        item = stmt.first_or_404()

        form = self.get_form()(obj=item, csrf_enabled=self.csrf_enabled)

        if form.validate_on_submit():

            for k, v in form.data.iteritems():
                setattr(item, k, v)

            item = item.update(commit=False, **kwargs)

            if self.model.has_property('user_id'):
                item.user_id = current_user.id

            item.save()

            if self.success_message:
                flash(self.success_message, category='success')

            return redirect(self.get_next_url(**kwargs))

        context = self.get_context()
        context.update({
            'form': form
        })
        return self.render(**context)


from flask import request
from flask_wtf import Form as BaseForm
from flask_wtf.form import _Auto
from wtforms import fields
from wtforms import validators
from wtforms.ext.csrf.fields import CSRFTokenField
from wtforms_json import flatten_json
import werkzeug.datastructures


class Form(BaseForm):

    def __init__(self, formdata=_Auto, *args, **kwargs):
        kwargs['csrf_enabled'] = False

        if formdata is _Auto and self.is_submitted() and request.json:
            formdata = flatten_json(self.__class__, request.json)
            formdata = werkzeug.datastructures.MultiDict(formdata)

        super(Form, self).__init__(formdata, *args, **kwargs)

    @classmethod
    def output_field_data_type(cls, field):
        if isinstance(field, (
                fields.HiddenField,
                fields.StringField,
                fields.TextField,
                fields.TextAreaField,
                fields.PasswordField,
                fields.RadioField,
                fields.SelectField,
                fields.SelectMultipleField,
        )):
            return 'str'

        if isinstance(field, (fields.IntegerField,)):
            return 'int'

        if isinstance(field, (fields.FloatField, fields.DecimalField,)):
            return 'float'

        if isinstance(field, (fields.DateField,)):
            return 'date'

        if isinstance(field, (fields.DateTimeField,)):
            return 'datetime'

        if isinstance(field, (fields.BooleanField,)):
            return 'bool'

        if isinstance(field, (fields.FileField,)):
            return 'file'

        if isinstance(field, (fields.FieldList,)):
            return 'list'

        if isinstance(field, (fields.FormField,)):
            return 'form'

        return 'str'

    @classmethod
    def output_validators(cls, field_validators):
        data = []
        for validator in field_validators:
            v_name = validator.__class__.__name__

            if isinstance(validator, validators.Length):
                content = '{} min:{} max:{}'.format(v_name, validator.min, validator.max)

            elif isinstance(validator, validators.Regexp):
                content = '{} regex:{}'.format(v_name, validator.regex.pattern)

            elif isinstance(validator, validators.EqualTo):
                content = '{} to:{}'.format(v_name, validator.fieldname)

            elif isinstance(validator, validators.NumberRange):
                content = '{} range:{}-{}'.format(v_name, validator.min, validator.max)

            elif isinstance(validator, validators.IPAddress):
                content = '{} type:{}'.format(v_name, 'ipv4' if validator.ipv4 else 'ipv6')

            else:
                content = v_name

            data.append(content)

        return data

    @classmethod
    def output_field(cls, field):
        data = dict()
        data['name'] = field.name
        data['text'] = field.label.text
        data['type'] = cls.output_field_data_type(field)

        if field.validators:
            data['validators'] = cls.output_validators(field.validators)

        if field.description:
            data['description'] = field.description

        if field.default is not None:
            data['default'] = field.default

        if isinstance(field, (fields.SelectField, fields.SelectMultipleField, fields.RadioField)):
            data['choices'] = field.choices

        if isinstance(field, (fields.FormField,)):
            data['form'] = cls.output_form(field.form_class(prefix=field.name))

        if isinstance(field, (fields.FieldList,)):
            if field.min_entries:
                data['min_entries'] = field.min_entries

            if field.max_entries:
                data['max_entries'] = field.max_entries

            name = '%s-{n}' % (field.short_name,)
            field = field.unbound_field.bind(form=None, name=name, _meta=field.meta)
            try:
                prefix = field.name + field.separator
                data['form'] = cls.output_form(field.form_class(prefix=prefix))
            except:
                data['field'] = cls.output_field(field)

        return data

    @classmethod
    def output_form(cls, form, **kwargs):
        data = []

        for field in form:
            if field.name in kwargs.keys():
                continue

            if isinstance(field, (fields.SubmitField,)):
                continue

            if not form.csrf_enabled and isinstance(field, (CSRFTokenField,)):
                continue

            data.append(cls.output_field(field))

        return data

    @classmethod
    def fields(cls, **kwargs):
        form = cls()
        return cls.output_form(form, **kwargs)


from setuptools import setup, find_packages

PACKAGE = "flask_zeus"
NAME = "flask-zeus"
DESCRIPTION = ""
AUTHOR = "nanlong"
AUTHOR_EMAIL = "fei.code@gmail.com"
URL = "https://github.com/nanlong/flask_zeus"
VERSION = __import__(PACKAGE).__version__

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license="BSD",
    url=URL,
    packages=find_packages(),
    platforms='python 3.5',
    install_requires=[
        'flask',
        'flask-login',
        'flask-wtf',
        'flask-sqlalchemy',
        'flask-restful',
        'wtforms-json',
    ]
)